function mk_model = mk_train(pos_train,neg_train,para,files)
% wrapper for multiplicative kernel training with SVMlight package 
% It assumes Gassian RBF kernel for K_\theta, and linear kernel for K_x,
% for a nonparametric model
% pos_train - each row is a training example for positive class (foreground)
% neg_train - each row is a training example for negative class (background)
% para - model parameters, para.eta for K_\theta, para.margin for stop criterion in bootstrap training, para.maxiter for max number of iterations. 
% mk_model - outputs from training, needed for reconstructing a family of detectors
% files.cla_file - file name to store mk_model, files.svml - dir for SVMlight package 
% files.tmp_file - temp file for SVMlight training

% data_dir = '/research/handshapes/Quan_Data/pedestrian_data/pami_data/';
% load('/research/handshapes/Quan_Data/pedestrian_data/pami_data/neg_train');
% load('/research/handshapes/Quan_Data/pedestrian_data/pami_data/neg_tough');
% neg_train = [neg_data{2}(1:3:end);neg_tough{2}];
% clear neg_data neg_tough;
% load('/research/handshapes/Quan_Data/pedestrian_data/pami_data/pos');
% pos_train = single(pos_data{2});
% %%%pos_masks = single(face_hog_mask(1:train_end,:));
% num_pos = size(pos_train,1);
% 
% cla_file = '/research/handshapes/Quan_Data/pedestrian_data/pami_data/PAS_mkcla';
addpath(files.svml,'-begin');
%%addpath('/research/handshapes/Quan_Data/shapeSeg/code/svm_light_para/svml/','-begin');

num_pos = size(pos_train,1);
num_neg = size(neg_train,1);
initial_flag = true;

neg_ini= num_pos;total_tough = num_pos;

%%%% bootstrap iterations 
margin = 0; iter = 0;
while margin < para.margin && iter <= para.maxiter
    iter = iter + 1;str = ['iteration = ',num2str(iter)]; disp(str);
    
    if initial_flag %%%% train the first round
        
        tough_ids = []; tough_fg_ids = []; %%% tough_fg_ids are indices of foreground examples whose associated detector has a false alarm
        initial_flag =false;
        num_neg = size(neg_train,1);
        neg_fg_ids = round(rand(neg_ini,1)*num_pos);      
        neg_fg_ids(neg_fg_ids == 0) = 1;
        pos_ids = (1:num_pos)';   neg_ids = (1:neg_ini)';
        %%%% obtain the k_\theta matrix
        %train_pos = pos_train.*pos_masks;
%         train_pos = train_pos.*repmat(5./sqrt(sum(train_pos.^2,2)),[1,size(train_pos,2)]);
 %       train_pos = pos_train;
        A = repmat(sum(pos_train.*pos_train,2),[1,size(pos_train,1)]);
        A_trans = A';   B = pos_train * pos_train';
        HRBF_dist = A + A' - B*2;
        %%%%% dist kernel %%%
        
        dist_exp = exp(-para.eta*HRBF_dist); clear HRBF_dist A_trans B A ;
        %%% a rule of thumb to choose eta is to make the median of dist_exp
        %%% around 0.2
    else
        %%load(cla_file);

        xsv_a = single(repmat(alphas,[1,size(xsv,2)]).*xsv);
        b_theta = single(ktheta*alphas);
        CLA = single(ktheta*xsv_a);

        b_theta_ps = repmat(b_theta,[1,size(pos_train,1)]);
        pos_scores = CLA*pos_train' + b_theta_ps - b;
        max_pos_scores = max(pos_scores);

        b_theta_m = repmat(b_theta,[1,size(neg_train,1)]);
        %%%neg_scores = CLA*neg_train' - b;
        neg_scores = CLA*neg_train' + b_theta_m - b;
        margin = min(max_pos_scores) - max(neg_scores(:));disp(['margin = ',num2str(margin)]);
        
        th = min(max_pos_scores);
        while 1

            %%% get hog from the  background training data %%%
            
            [claids, hogids] = find(neg_scores > th);%%%length(claids)
            mids = sub2ind(size(neg_scores),claids,hogids);
            if length(claids) < total_tough
                th = th - 0.1;continue;
            end
            [backsort,bsid] = sort(neg_scores(mids),'descend'); %%%%find largest scores of false postives
            sample_ids = bsid(1:total_tough);
%             steps = length(claids)/total_tough;
%             sample_ids = round(1:steps:length(claids)); %%% sample along cla_ids (mask_ids)
            tough_fg_ids = claids(sample_ids);
            tough_ids = hogids(sample_ids);

            break;
        end
     
    end
    neg_ids = [neg_ids;tough_ids]; neg_fg_ids = [neg_fg_ids;tough_fg_ids];
    %%% neg_train_ids = [neg_ids;tough_ids]; neg_mask_train_ids = [neg_mask_ids;tough_mask_ids];
    %%%% id_v, pos_ids, neg_fg_ids,tough_fg_ids are ids in pos_train, neg_ids
    %%%% is for neg_train
    id_v = [pos_ids;neg_fg_ids];  %%% i in all (x,i) tuples  
    num_all = length(id_v);
    labels = [ones(num_pos,1);-ones(num_all - num_pos,1)];
    %%% replace id_v with new id_v , which are assoicated positive sample indices %%
    
    xx = [pos_train;neg_train(neg_ids,:)]; %%% x in all (x,i) tuples 
    xx = xx.*repmat(5./sqrt(sum(xx.^2,2)),[1,size(xx,2)]);
    num_all = length(id_v);
    id_matrix = zeros(num_all,num_all,2);
    id_matrix(:,:,1) = uint16(repmat(id_v,[1,num_all]));
    id_matrix(:,:,2) = uint16(squeeze(id_matrix(:,:,1))');
   
    ind_matrix =(id_matrix(:,:,2)-1)*size(dist_exp,1) + id_matrix(:,:,1);
    
    clear id_matrix; K_theta = dist_exp(ind_matrix);
    clear ind_matrix; K_x = xx*xx'+1; K_gram = K_theta.*K_x;
    clear K_x; K_gram_ext = [(1:num_all)',K_gram]; clear K_gram
    
    fname = files.tmp_file;
    
    Net = svml(fname,'ExecPath',files.svml,'C',1,'Kernel',4,'KernelParam',10,'Verbosity',0);
    %%Net = svml(fname,'C',1,'Kernel',4,'KernelParam',10,'Verbosity',0);
    Net = svmltrain(Net,K_gram_ext,labels); clear K_gram_ext;
   
    [svs,alphas,b,rbf_gamma] = mexLoadSVMLightSVs(fname);
    v_ids = svs(:,1); clear svs;
    xsv = xx(v_ids,:);clear xx;
    ktheta = K_theta(1:num_pos,v_ids); clear K_theta;
    %%%xsv_a = repmat(alphas,[1,size(xsv,2)]).*xsv;
    
    %%%%% only keep sv neg ids for next round
    %%%total_ids = [pos_ids;neg_train_ids]; 
    %%%neg_curr_ids = total_ids(alphas<0); neg_curr_mask_ids = id_v(alphas < 0);
    %%neg_ids = [neg_ids;neg_curr_ids]; neg_mask_ids = [neg_mask_ids;neg_curr_mask_ids];
    %ped_model.v_ids = v_ids; model.alphas = alphas; model.b = b; model.ktheta = ktheta; model.xsv = xsv; model.id_v = id_v;
    %save(files.cla_file,'v_ids','alphas','b','rbf_gamma','dist_exp','num_pos','ktheta','xsv','id_v','neg_ids','neg_mask_ids','eta','-v7.3');
end
mk_model.v_ids = v_ids; 
mk_model.alphas = alphas; 
mk_model.b = b; 
mk_model.ktheta = ktheta; 
mk_model.xsv = xsv; 
mk_model.id_v = id_v;
   

