/************************************************************************/
/*                                                                      */
/*   kernel.h                                                           */
/*                                                                      */
/*   User defined kernel function. Feel free to plug in your own.       */
/*                                                                      */
/*   Copyright: Thorsten Joachims                                       */
/*   Date: 16.12.97                                                     */
/*                                                                      */
/************************************************************************/

/* KERNEL_PARM is defined in svm_common.h The field 'custom' is reserved for */
/* parameters of the user defined kernel. You can also access and use */
/* the parameters of the other kernels. Just replace the line 
             return((double)(1.0)); 
   with your own kernel. */

  /* Example: The following computes the polynomial kernel. sprod_ss
              computes the inner product between two sparse vectors. 

      return((CFLOAT)pow(kernel_parm->coef_lin*sprod_ss(a->words,b->words)
             +kernel_parm->coef_const,(double)kernel_parm->poly_degree)); 
  */

/* If you are implementing a kernel that is not based on a
   feature/value representation, you might want to make use of the
   field "userdefined" in SVECTOR. By default, this field will contain
   whatever string you put behind a # sign in the example file. So, if
   a line in your training file looks like

   -1 1:3 5:6 #abcdefg

   then the SVECTOR field "words" will contain the vector 1:3 5:6, and
   "userdefined" will contain the string "abcdefg". */
//#include "mex.h"
# include "svm_common.h"


double RBF_linear_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b);
double RBF_RBF_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b);
double matrix_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b);

double custom_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b) 
     /* plug in you favorite kernel */                        
{
    
	int i;
    //for(i = 0;i < a->n_words;i++) printf("a[%d] =  %f  ",i, a->words[i]); printf("\n");
    //for(i = 0;i < b->n_words;i++) printf("b[%d] =  %f  ",i, b->words[i]); printf("\n");
    return matrix_kernel(kernel_parm, a, b);
    //return RBF_linear_kernel(kernel_parm, a, b);
	//return RBF_RBF_kernel(kernel_parm, a, b);
}

double matrix_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b)
{// the input is the GRAM matrix + the first element is the index of the sample    
 // the input come from Mablab, assuming the index starting from 1 !!
	int i;
	i = a->words[0];
	if(a->words[0] < 1)
	{ //printf("has 0 index! \n"); 
	  return 0;
	}
	else
       return(b->words[i]);
  
}


double RBF_linear_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b)
{// each row is formated as [num_parameters, parameter_vector, feature_vector]   
  int num_paras = a->words[0];
  double sigma = kernel_parm->rbf_gamma;
  double kernel_theta, theta_d = 0,d_prod = 0;
  int i;
  for(i = 1;i <= num_paras; i++ )//calculate the parameter kernel
  {
      //printf("a->words = %f, b->words = %f \n", a->words[i],b->words[i]);
	  theta_d += a->words[i] * b->words[i];
  }
  kernel_theta = exp(-kernel_parm->rbf_gamma*(a->twonorm_sq-2*theta_d+b->twonorm_sq));
  
	  // calculate the feature kernel
  for(i = num_paras+1;i < a->n_words; i++ )
         d_prod += a->words[i] *b->words[i];
  //printf("num_feature = %d, d_prod = %f, kernel_theta = %f, sigma = %f, num_paras = %d, theta_d = %f\n", a->n_words,d_prod,kernel_theta,sigma, num_paras, theta_d);
  return(kernel_theta*d_prod  + kernel_theta); 
}

double RBF_RBF_kernel(KERNEL_PARM *kernel_parm, SVECTOR *a, SVECTOR *b)
{ // each row is formated as [num_parameters, parameter_vector, feature_gamma, feature_vector]   
  int num_paras = a->words[0];
  double sigma = kernel_parm->rbf_gamma;
  double kernel_theta,kernel_x, theta_d = 0,d_prod = 0;
  int i;
  for(i = 1;i <= num_paras; i++ )//calculate the parameter kernel
	  theta_d += a->words[i] * b->words[i];
 
  kernel_theta = exp(-kernel_parm->rbf_gamma*(a->twonorm_sq-2*theta_d+b->twonorm_sq));
  
	  // calculate the feature kernel
  for(i = num_paras+2;i < a->n_words; i++ )
         d_prod += a->words[i] *b->words[i];
  kernel_x = exp(-a->words[num_paras+1]*(a->twonorm_sq2-2*d_prod+b->twonorm_sq2));
  //printf("num_feature = %d, d_prod = %f, kernel_theta = %f, sigma = %f, num_paras = %d, theta_d = %f,kernel_x = %f,gamma = %f \n", a->n_words,d_prod,kernel_theta,sigma, num_paras, theta_d,kernel_x,a->words[num_paras+1]);
  //for(i = 0;i < a->n_words;i++) printf("a[%d] =  %f  ",i, a->words[i]); printf("\n");
  //for(i = 0;i < b->n_words;i++) printf("b[%d] =  %f  ",i, b->words[i]); printf("\n");
  return(kernel_theta*kernel_x ); 
}
